#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2017, Ilya Etingof <etingof@gmail.com>
# License: http://pysnmp.sf.net/license.html
#
from pysnmp.proto import rfc1905

noSuchObject = rfc1905.noSuchObject
noSuchInstance = rfc1905.noSuchInstance
endOfMibView = endOfMib = rfc1905.endOfMibView
